import os
from datetime import datetime
import shutil
import numpy as np
from matplotlib import pyplot as plt
from pandas import DataFrame
import matplotlib.lines as mlines
import pandas
import matplotlib
from netCDF4 import Dataset

#expids = ['a327','a328','a329','a33u','a33v']
#expids = ['a27e','a27f','a27g','a27m','a274']
expids = ['a27e']
templatepath = '/gpfs/scratch/bsc32/bsc32786/pisces1d/parameters/variables5wchld.csv'
mode = 'obs' # 'obs' or 'def'
scr = '/gpfs/scratch/bsc32/bsc32786/'
modelfile='PISCES_5d_19600101_19691231_ptrc_T.nc'

obsfile= scr + 'pisces1d/obs/Mprof_6901486_L75_5days_2015-01-01_2015-12-31reg.nc' # no need to specify in mode = def

#### create new directory, and copy the results into it


os.chdir(os.getcwd()+'/logs')
pwd = os.getcwd() + '/'
dt = datetime.now()
d = str(dt.year)[-2:]+'-'+"{:02d}".format(dt.month)+'-'+"{:02d}".format(dt.day)
procs = [int(i[-1]) for i in os.listdir(pwd) if d in i ]
procid = 0

if len(procs) > 0:
    while procid in procs:
        procid = procid+1
procid=str(procid)
os.mkdir(pwd + d + '_' + procid)
os.chdir(pwd + d + '_' + procid)
pwd = os.getcwd() + '/'

shutil.copyfile(templatepath, pwd)

outputfile = 'PISCES_5d_19600101_19641231_ptrc_T.nc'

for expid in expids:
    os.chdir(pwd)
    if expid in os.listdir(os.getcwd()):
        shutil.rmtree(pwd + expid)
    os.mkdir(pwd + expid)
    os.chdir(pwd + expid)

    os.mkdir('parameters')
    os.chdir('parameters')
    fs = os.listdir(scr + expid + '/EXP00/parameters/')
    for fi in fs:
        shutil.copyfile(scr + expid + '/EXP00/parameters/' + fi, os.getcwd() + '/' + fi)
    os.remove('0000.csv')

    os.chdir(pwd + expid)
    os.mkdir('stats')
    os.chdir('stats')
    fs = os.listdir(scr + expid + '/EXP00/bPOC/')
    for fi in fs:
        shutil.copyfile(scr + expid + '/EXP00/bPOC/' + fi, os.getcwd() + '/' + fi)
    fs = os.listdir(scr + expid + '/EXP00/sPOC/')
    for fi in fs:
        shutil.copyfile(scr + expid + '/EXP00/sPOC/' + fi, os.getcwd() + '/' + fi)

    # obtain highest performing file

    os.chdir(pwd + expid)
    fs = sorted(os.listdir(pwd + expid + '/parameters'))
    paras = np.genfromtxt(pwd + expid + '/parameters/'+fs[-1])
    minind = np.where(paras[:, -1] == min(paras[:, -1]))[0][0]
    chunk = "{0:0=4d}".format(minind+1)
    shutil.copyfile(scr + expid + '/EXP00/'+ chunk + '/' + outputfile , os.getcwd() + '/' + outputfile + expid)
    shutil.copyfile(scr + expid + '/EXP00/default/' + outputfile, os.getcwd() + '/' + outputfile + 'def')

### plot the evolution of parameters and save to a table


os.chdir('..')

#os.chdir('/esarchive/scratch/mfalls/pisces1d/logs/21-01-20_0')

def make_grid(n):
    y=1
    x=1
    while y*x < n:
        if y==x:
            y=y+1
        else:
            x=x+1
    j=0
    i=0
    tuples=[]
    for a in range(n):
        tuples.append((j,i))
        if i == x - 1:
            i=0
            j=j+1
        else:
            i=i+1

    return y,x,tuples

runs = expids

basepath=os.getcwd() + '/'
paraspath='/parameters/'
bbp700path='/bbp700/'

template = pandas.read_csv(templatepath)
parameters=list(template['var'])
defaults=list(template['defaults'])
mins=list(template['min'])
maxes=list(template['max'])
numparams = len(parameters)


Y,X,T = make_grid(numparams)
print(Y,X,T)

cols=['C0','C1','C2','C3','C4','C5','C6']

fig_size = plt.rcParams["figure.figsize"]
fig_size[0] = 12.8
fig_size[1] = 9.6

plt.rcParams["figure.figsize"] = fig_size
plt.locator_params(nbins=6)

fig, axes = plt.subplots(Y, X, sharex=True)
fig.text(0.5, 0.04, 'Generation', ha='center', size='large')

costs={}

for r in range(len(runs)):
    costs[runs[r]]=[]
    stats = []
    for a in range(numparams):
        stats.append([])
    parafilelst = os.listdir(basepath + runs[r] + paraspath)
    parafilelst.sort()

    parasDF = DataFrame(index=range(len(parafilelst)), columns=parameters)

    for f in parafilelst:
        paras = np.genfromtxt(basepath + runs[r] + paraspath + f)
        minind = np.where(paras[:, -1] == min(paras[:, -1]))
        costs[runs[r]].append(float(paras[minind, -1][0][0]))
        for a in range(numparams):
            stats[a].append(paras[minind][0][a])

    for a in range(numparams):
        axes[T[a][0], T[a][1]].plot(range(len(stats[a])), stats[a], color=cols[r])#, color='grey', ls='--', linewidth=1)
        parasDF[parameters[a]]=stats[a]

    parasDF.to_csv('parameters_'+runs[r]+'.csv')

for a in range(numparams):
    axes[T[a][0],T[a][1]].plot( range(len(costs[runs[0]])),  [defaults[a]]*len(costs[runs[0]]), color='grey', ls='--', linewidth=1)
    axes[T[a][0],T[a][1]].plot( range(len(costs[runs[0]])),  [mins[a]]*len(costs[runs[0]]), color='grey', ls=':', linewidth=1)
    axes[T[a][0],T[a][1]].plot( range(len(costs[runs[0]])),  [maxes[a]]*len(costs[runs[0]]), color='grey', ls=':', linewidth=1)
    axes[T[a][0], T[a][1]].set_title(parameters[a])

legs=[]

legs.append(mlines.Line2D([], [], color='gray', ls='--', label='Default'))
for r in range(len(runs)):
    legs.append(mlines.Line2D([], [], color=cols[r], ls='-', label=runs[r]))
leg1=fig.legend(handles=legs, loc='center right', bbox_to_anchor=(1, 0.5))
plt.gca().add_artist(leg1)

for j in range(len(axes[:,0])):
    for i in range(len(axes[0,:])):
        if not axes[j,i].lines:
            fig.delaxes(axes[j][i])

plt.savefig('parameters.png')
plt.cla()


## plot the evolution of costs and statistics


fig_size = plt.rcParams["figure.figsize"]
fig_size[0] = 6.4
fig_size[1] = 4.8
plt.rcParams["figure.figsize"] = fig_size

fig = plt.figure()


costsDF=DataFrame.from_dict(costs,orient='index').transpose()
costsDF.to_csv('costs.csv')

for r in runs:
    plt.plot( range(len(costs[r])), costs[r])

if mode == 'obs':
    defcost = np.genfromtxt(runs[0] + '/parameters/0001.csv')[0,-1]
    print(defcost)
    plt.plot(range(len(costs[r])), [defcost]*len(costs[runs[0]]), color='grey', ls='--', linewidth=1)

legs=[]

legs.append(mlines.Line2D([], [], color='gray', ls='--', label='Default'))
for r in range(len(runs)):
    legs.append(mlines.Line2D([], [], color=cols[r], ls='-', label=runs[r]))
leg1=fig.legend(handles=legs, loc='center right', bbox_to_anchor=(1, 0.5))
plt.gca().add_artist(leg1)
plt.xlabel('Generation')
plt.ylabel('Cost (ST Score)')
plt.savefig('costs.png')

statnames = ['sPOC', 'bPOC']
for statname in statnames:
    fig, axes = plt.subplots(2, 2, sharex=True)
    fig.text(0.5, 0.04, 'Generation', ha='center', size='medium')
    costs={}

    for r in range(len(runs)):

        filelst=sorted([i for i in os.listdir(basepath + runs[r] + '/stats/') if statname in i])

        allbias=[]
        allstd=[]
        allr=[]
        allrmse=[]
        for f in range(len(filelst)):
            data = pandas.read_csv(basepath + runs[r] + '/stats/' + filelst[f])
            #print(data)
            allbias.append(data['bias'][0])
            allstd.append(data['normstd'][0])
            allr.append(data['R'][0])
            allrmse.append(data['RMSE'][0])

        axes[0,0].plot(allbias, color=cols[r])
        axes[0, 1].plot(allstd, color=cols[r])
        axes[1, 0].plot(allr, color=cols[r])
        axes[1, 1].plot(allrmse, color=cols[r])

        statsDF = DataFrame(index=range(len(filelst)), columns=['bias','normstd','R','RMSE'])
        statsDF['bias']=allbias
        statsDF['normstd']=allstd
        statsDF['R']=allr
        statsDF['RMSE']=allrmse
        statsDF.to_csv(statname + '_stats_' + runs[r] + '.csv')

    legs=[]

    legs.append(mlines.Line2D([], [], color='gray', ls='--', label='Default'))
    for r in range(len(runs)):
        legs.append(mlines.Line2D([], [], color=cols[r], ls='-', label=runs[r]))
    leg1=fig.legend(handles=legs, loc='center right', bbox_to_anchor=(1, 0.5))
    plt.gca().add_artist(leg1)

    axes[ 0, 0].set_title('Bias',size='medium')
    axes[ 0, 1].set_title('Standard Deviation',size='medium')
    axes[ 1, 0].set_title('Correlation',size='medium')
    axes[ 1, 1].set_title('RMSE',size='medium')

    if mode == 'obs':
        defstats = pandas.read_csv(runs[0] + '/stats/'+statname+'0001.csv')
        axes[ 0, 0].plot(range(len(allbias)), [defstats['bias'][0]] * len(allbias), color='grey', ls='--', linewidth=1)
        axes[0, 1].plot(range(len(allbias)), [defstats['normstd'][0]] * len(allbias), color='grey', ls='--', linewidth=1)
        axes[1, 0].plot(range(len(allbias)), [defstats['R'][0]] * len(allbias), color='grey', ls='--', linewidth=1)
        axes[1, 1].plot(range(len(allbias)), [defstats['RMSE'][0]] * len(allbias), color='grey', ls='--', linewidth=1)

    plt.savefig(statname+'_stats.png')

##### plot the actual pocs

if mode == 'obs':
    OBS = Dataset(obsfile)
    DEF = Dataset(expids[0]+'/PISCES_5d_19600101_19641231_ptrc_T.ncdef')
    OPT = Dataset(expids[0]+'/PISCES_5d_19600101_19641231_ptrc_T.nc'+expids[0])


    def getobsdata(A):
        A[A == 0.0] = np.nan
        A = np.log10(A)
        mn = np.nanmean(A)
        std = np.nanstd(A)
        return A, mn, std


    def editmoddata(arr, d, s, mn=None, std=None):
        arr = arr[:, :np.shape(d)[0]]
        arr[-s:]
        arr[arr == 200000039192547557376] = np.nan
        arr = np.log10(arr)
        if mn:
            return arr
        else:
            mn = np.nanmean(arr)
            std = np.nanstd(arr)
            return arr, mn, std


    sPOCobs = np.array(OBS.variables['spoc'])
    sPOCobs = sPOCobs[:, :]
    sPOCobs, sPOCobsmean, sPOCobsstd = getobsdata(sPOCobs)
    depths = np.array(OBS.variables['deptht'])
    S = np.shape(sPOCobs)[0]

    sPOCdef = (np.array(DEF.variables['POC'])[-S:, :, 1, 1] + np.array(DEF.variables['PHY'])[-S:, :, 1, 1] + np.array(
        DEF.variables['ZOO'])[-S:, :, 1, 1])
    sPOCdef = sPOCdef[:, :]
    sPOCdef = editmoddata(sPOCdef, depths, S, mn=sPOCobsmean, std=sPOCobsstd)

    sPOCopt = (np.array(OPT.variables['POC'])[-S:, :, 1, 1] + np.array(OPT.variables['PHY'])[-S:, :, 1, 1] + np.array(
        OPT.variables['ZOO'])[-S:, :, 1, 1])
    sPOCopt = sPOCopt[:, :]
    sPOCopt = editmoddata(sPOCopt, depths, S, mn=sPOCobsmean, std=sPOCobsstd)

    sPOCobs[np.isnan(sPOCobs)] = 0
    sPOCobs[sPOCobs == 0] = np.nan

    current_cmap = matplotlib.cm.get_cmap()
    current_cmap.set_bad(color='white')

    spocall = np.array([sPOCobs, sPOCdef, sPOCopt])
    min1, max1 = np.nanmin(spocall), np.nanmax(spocall)

    yticks = [int(i) for i in list(depths[:])]

    fig = plt.figure()

    xticks = ['Feb', 'May', 'Aug', 'Nov']
    yinds = range(0, len(depths), 5)
    yticks = [int(i) for i in list(depths[yinds])]

    ax1 = fig.add_subplot(231)
    plt.imshow(sPOCobs.T, vmin=min1, vmax=max1)  # , extent = [0,1,depths[-1],depths[0]])

    ax1.set_aspect(2)
    ax1.set_xticks([9, 27, 45, 63])
    ax1.set_xticklabels(xticks, fontsize=5)
    ax1.set_yticks(yinds)
    ax1.set_yticklabels(yticks, fontsize=5)
    ax1.set_ylabel('Depth (m)', fontsize=8)
    ax1.set_xlabel('Observed', fontsize=8)

    ax2 = fig.add_subplot(232)
    plt.title('sPOC')
    plt.imshow(sPOCdef.T, vmin=min1, vmax=max1)
    ax2.set_aspect(2)
    ax2.set_xticks([9, 27, 45, 63])
    ax2.set_xticklabels(xticks, fontsize=5)
    ax2.set_yticks(yinds)
    ax2.set_yticklabels([])
    ax2.set_xlabel('Default', fontsize=8)

    ax3 = fig.add_subplot(233)
    plt.imshow(sPOCopt.T, vmin=min1, vmax=max1)
    ax3.set_aspect(2)
    ax3.set_xticks([9, 27, 45, 63])
    ax3.set_xticklabels(xticks, fontsize=5)
    ax3.set_yticks(yinds)
    ax3.set_yticklabels([])
    ax3.set_xlabel('Optimised', fontsize=8)

    sPOCdefbias = sPOCdef.T - sPOCobs.T
    sPOCoptbias = sPOCopt.T - sPOCobs.T

    sPOCbiasall = np.array([sPOCdefbias, sPOCoptbias])

    min2, max2 = np.nanmin(sPOCbiasall), np.nanmax(sPOCbiasall)
    if abs(max2) > abs(min2):
        min2 = -max2
    else:
        max2 = -min2

    ax4 = fig.add_subplot(235)
    plt.imshow(sPOCdefbias, cmap=plt.cm.seismic, vmin=min2, vmax=max2)
    ax4.set_aspect(2)
    ax4.set_xticks([9, 27, 45, 63])
    ax4.set_xticklabels(xticks, fontsize=5)
    ax4.set_yticks(yinds)
    ax4.set_yticklabels(yticks, fontsize=5)
    ax4.set_ylabel('Depth (m)', fontsize=8)
    ax4.set_xlabel('Default bias', fontsize=8)

    ax5 = fig.add_subplot(236)
    plt.imshow(sPOCoptbias, cmap=plt.cm.seismic, vmin=min2, vmax=max2)
    ax5.set_aspect(2)
    ax5.set_xticks([9, 27, 45, 63])
    ax5.set_xticklabels(xticks, fontsize=5)
    ax5.set_yticks(yinds)
    ax5.set_yticklabels([])
    ax5.set_xlabel('Optimised bias', fontsize=8)

    fig.savefig('sPOC.png')

    print('sPOC.png saved')
    ### bpoc

    plt.cla()

    bPOCobs = np.array(OBS.variables['bpoc'])
    bPOCobs = bPOCobs[:, :]
    bPOCobs, bPOCobsmean, bPOCobsstd = getobsdata(bPOCobs)
    depths = np.array(OBS.variables['deptht'])
    S = np.shape(bPOCobs)[0]

    bPOCdef = (np.array(DEF.variables['GOC'])[-S:, :, 1, 1] + np.array(DEF.variables['PHY2'])[-S:, :, 1, 1] + np.array(
        DEF.variables['ZOO2'])[-S:, :, 1, 1])
    bPOCdef = bPOCdef[:, :]
    bPOCdef = editmoddata(bPOCdef, depths, S, mn=bPOCobsmean, std=bPOCobsstd)

    bPOCopt = (np.array(OPT.variables['GOC'])[-S:, :, 1, 1] + np.array(OPT.variables['PHY2'])[-S:, :, 1, 1] + np.array(
        OPT.variables['ZOO2'])[-S:, :, 1, 1])
    bPOCopt = bPOCopt[:, :]
    bPOCopt = editmoddata(bPOCopt, depths, S, mn=bPOCobsmean, std=bPOCobsstd)

    bPOCobs[np.isnan(bPOCobs)] = 0
    bPOCobs[bPOCobs == 0] = np.nan

    current_cmap = matplotlib.cm.get_cmap()
    current_cmap.set_bad(color='white')

    bpocall = np.array([bPOCobs, bPOCdef, bPOCopt])
    min1, max1 = np.nanmin(bpocall), np.nanmax(bpocall)

    yticks = [int(i) for i in list(depths[:])]

    fig = plt.figure()

    xticks = ['Feb', 'May', 'Aug', 'Nov']
    yinds = range(0, len(depths), 5)
    yticks = [int(i) for i in list(depths[yinds])]

    ax1 = fig.add_subplot(231)
    plt.imshow(bPOCobs.T, vmin=min1, vmax=max1)  # , extent = [0,1,depths[-1],depths[0]])

    ax1.set_aspect(2)
    ax1.set_xticks([9, 27, 45, 63])
    ax1.set_xticklabels(xticks, fontsize=5)
    ax1.set_yticks(yinds)
    ax1.set_yticklabels(yticks, fontsize=5)
    ax1.set_ylabel('Depth (m)', fontsize=8)
    ax1.set_xlabel('Observed', fontsize=8)

    ax2 = fig.add_subplot(232)
    plt.title('bPOC')
    plt.imshow(bPOCdef.T, vmin=min1, vmax=max1)
    ax2.set_aspect(2)
    ax2.set_xticks([9, 27, 45, 63])
    ax2.set_xticklabels(xticks, fontsize=5)
    ax2.set_yticks(yinds)
    ax2.set_yticklabels([])
    ax2.set_xlabel('Default', fontsize=8)

    ax3 = fig.add_subplot(233)
    plt.imshow(bPOCopt.T, vmin=min1, vmax=max1)
    ax3.set_aspect(2)
    ax3.set_xticks([9, 27, 45, 63])
    ax3.set_xticklabels(xticks, fontsize=5)
    ax3.set_yticks(yinds)
    ax3.set_yticklabels([])
    ax3.set_xlabel('Optimised', fontsize=8)

    bPOCdefbias = bPOCdef.T - bPOCobs.T
    bPOCoptbias = bPOCopt.T - bPOCobs.T

    bPOCbiasall = np.array([bPOCdefbias, bPOCoptbias])

    min2, max2 = np.nanmin(bPOCbiasall), np.nanmax(bPOCbiasall)
    if abs(max2) > abs(min2):
        min2 = -max2
    else:
        max2 = -min2

    ax4 = fig.add_subplot(235)
    plt.imshow(bPOCdefbias, cmap=plt.cm.seismic, vmin=min2, vmax=max2)
    ax4.set_aspect(2)
    ax4.set_xticks([9, 27, 45, 63])
    ax4.set_xticklabels(xticks, fontsize=5)
    ax4.set_yticks(yinds)
    ax4.set_yticklabels(yticks, fontsize=5)
    ax4.set_ylabel('Depth (m)', fontsize=8)
    ax4.set_xlabel('Default bias', fontsize=8)

    ax5 = fig.add_subplot(236)
    plt.imshow(bPOCoptbias, cmap=plt.cm.seismic, vmin=min2, vmax=max2)
    ax5.set_aspect(2)
    ax5.set_xticks([9, 27, 45, 63])
    ax5.set_xticklabels(xticks, fontsize=5)
    ax5.set_yticks(yinds)
    ax5.set_yticklabels([])
    ax5.set_xlabel('Optimised bias', fontsize=8)

    fig.savefig('bPOC.png')

    print('bPOC.png saved')
